@@ -2,12 +2,10 @@
 
    \brief RLog logging routines
 
-   $Id: rlog.cxx,v 1.8 2008-11-21 21:16:45 burghardt Exp $
-
 *//*
 
    ClamFS - An user-space anti-virus protected file system
-   Copyright (C) 2007 Krzysztof Burghardt.
+   Copyright (C) 2007-2019 Krzysztof Burghardt
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,6 +26,8 @@
 
 #include "rlog.hxx"
 
+#include <unistd.h>
+
 namespace clamfs {
 
 extern config_t config;
@@ -94,7 +94,7 @@ void RLogOpenLogFile(const char *filename) {
     mode_t mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
 
     fileLog = open(filename, O_WRONLY | O_CREAT | O_APPEND, mode);
-    if (fileLog > 0) { /* file open succesful */
+    if (fileLog > 0) { /* file open successful */
 #ifndef NDEBUG
         fileLogNode = new StdioNode(fileLog, StdioNode::OutputContext |
         StdioNode::OutputThreadId);
