@@ -2,12 +2,10 @@
 
    \brief ScanCache (anti-virus scan result caching) routines
 
-   $Id: scancache.cxx,v 1.7 2008-11-23 20:50:00 burghardt Exp $
-
 *//*
 
    ClamFS - An user-space anti-virus protected file system
-   Copyright (C) 2007,2008 Krzysztof Burghardt.
+   Copyright (C) 2007-2019 Krzysztof Burghardt
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -37,7 +35,7 @@ CachedResult::~CachedResult() {
 }
 
 ScanCache::ScanCache(long int elements, long int expire):
-    ExpireLRUCache<ino_t, CachedResult>::ExpireLRUCache<ino_t, CachedResult>(elements, expire) {
+    ExpireLRUCache<ino_t, CachedResult>(elements, expire) {
 }
 
 ScanCache::~ScanCache() {
